@@ -33,6 +33,7 @@ from collections import Counter
 from typing import Tuple, Optional, Iterable, Iterator, List, TYPE_CHECKING, Union
 from sortedcontainers import SortedSet
 
+import numba as nb
 import numpy as np
 
 from .dissimilarity import AbstractDissimilarity, CombinedCategoricalDissimilarity
@@ -181,16 +182,27 @@ class Alignment(AbstractAlignment):
     def __iter__(self) -> Iterator[UnitaryAlignment]:
         return iter(self.unitary_alignments)
 
-    @property
-    def leftmost(self):
-        """Return the (or one of the) leftmost unitary alignments."""
-        return min(self.unitary_alignments, key=(lambda unitary_alignment: unitary_alignment.bounds[0]))
+    def take_until_limit(self, x_limit):
+        for unitary_alignment in sorted(self.unitary_alignments, key=lambda unit_align: unit_align.bounds[1]):
+            if unitary_alignment.bounds[1] > x_limit:
+                break
+            yield unitary_alignment
 
 
     @property
     def annotators(self):
         return SortedSet([annotator for annotator, _
-                in self.unitary_alignments[0].n_tuple])
+                          in self.unitary_alignments[0].n_tuple])
+
+    @property
+    def categories(self):
+        if self.continuum is not None:
+            return self.continuum.categories
+        else:
+            return SortedSet([unit.annotation
+                              for unitary_alignment in self
+                              for _, unit in unitary_alignment.n_tuple
+                              if unit is not None])
 
     @property
     def avg_num_annotations_per_annotator(self):
@@ -208,8 +220,14 @@ class Alignment(AbstractAlignment):
         return len(self.unitary_alignments[0].n_tuple)
 
     @property
-    def disorder(self):
-        # TODO : doc
+    def disorder(self) -> float:
+        """
+
+        Returns
+        -------
+        float:
+            The disorder of the alignment.
+        """
         if self._disorder is None:
             self._disorder = (sum(u_align.disorder for u_align
                                   in self.unitary_alignments)
@@ -219,17 +237,12 @@ class Alignment(AbstractAlignment):
     def compute_disorder(self, dissimilarity: AbstractDissimilarity):
         """
         Recalculates the disorder of this alignment using the given dissimilarity computer.
-        Usually not needed since most alignment are generated from a minimal disorder
-        so they are
+        Usually not needed since most alignment are generated from a minimal disorder.
         """
-        disorder_args = dissimilarity.build_args(self)
-        unit_ids = np.arange(self.num_unitary_alignments, dtype=np.int32)
-        unit_ids = np.vstack([unit_ids] * self.num_annotators)
-        unit_ids = unit_ids.swapaxes(0, 1)
-        disorders = dissimilarity(unit_ids, *disorder_args)
+        disorders = dissimilarity.compute_disorder(self)
         for i, disorder in enumerate(disorders):
             self.unitary_alignments[i].disorder = disorder
-        self._disorder = (dissimilarity(unit_ids, *disorder_args).sum()
+        self._disorder = (np.sum(disorders)
                           / self.avg_num_annotations_per_annotator)
         return self._disorder
 
@@ -263,7 +276,7 @@ class Alignment(AbstractAlignment):
             else:
                 weight_base = 1 / (nv - 1)
             for i, (_, unit1) in enumerate(unitary_alignment.n_tuple):
-                for _, unit2 in unitary_alignment.n_tuple[i+1:]:
+                for _, unit2 in unitary_alignment.n_tuple[i + 1:]:
                     # Case handler for gamma-k
                     if category is not None and ((unit1 is None or unit1.annotation != category)
                                                  and (unit2 is None or unit2.annotation != category)):
@@ -280,8 +293,8 @@ class Alignment(AbstractAlignment):
                     weight_confidence = max(0, 1 - pos_dissim)
                     cat_dissim = dissimilarity.categorical_dissim.d(unit1, unit2)
                     weight = weight_base * weight_confidence  # Each categorical dissimilarity is weighted by both
-                    total_disorder += cat_dissim * weight     # a positional "confidence" and the # of alignments
-                    total_weight += weight                    # in the unitary alignment
+                    total_disorder += cat_dissim * weight  # a positional "confidence" and the # of alignments
+                    total_weight += weight  # in the unitary alignment
         if no_loop:
             return 1.0 if no_cat else 0.0
         return 0 if total_disorder == 0 else total_disorder / total_weight
@@ -357,3 +370,5 @@ class Alignment(AbstractAlignment):
 
         from .notebook import repr_alignment
         return repr_alignment(self)
+
+
