@@ -26,22 +26,19 @@
 ##########
 Continuum and corpus
 ##########
-
 """
 import csv
 import logging
 import os
-import sys
-import time
 from copy import deepcopy
 from dataclasses import dataclass
 from functools import total_ordering
-from multiprocessing import Pool
+from concurrent.futures import ThreadPoolExecutor
 from pathlib import Path
-from itertools import islice
-from typing import Optional, Tuple, List, Union, Set, Iterable, TYPE_CHECKING, Dict, Generator
+from typing import Optional, Tuple, List, Union, TYPE_CHECKING, Generator
 
 import cvxpy as cp
+import matplotlib.pyplot as plt
 import numpy as np
 from pyannote.core import Annotation, Segment, Timeline
 from pyannote.database.util import load_rttm
@@ -49,7 +46,6 @@ from sortedcontainers import SortedDict, SortedSet
 from typing_extensions import Literal
 
 from .dissimilarity import AbstractDissimilarity
-from .numba_utils import chunked_cartesian_product
 
 if TYPE_CHECKING:
     from .alignment import UnitaryAlignment, Alignment
@@ -70,6 +66,7 @@ PRECISION_LEVEL = {
 }
 
 
+
 @total_ordering
 @dataclass(frozen=True, eq=True)
 class Unit:
@@ -103,7 +100,7 @@ class Unit:
 class Continuum:
     """
     Representation of a continuum, i.e a set of annotated segments by multiple annotators.
-    It is implemented as a dictionnarfrom .notebook import show_continuumy of sets (all sorted) :
+    It is implemented as a dictionnary of sets (all sorted) :
 
     ``{'annotator1': {unit1, ...}, ...}``
     """
@@ -124,9 +121,13 @@ class Continuum:
         self.uri = uri
         # Structure {annotator -> SortedSet}
         self._annotations: SortedDict = SortedDict()
+        self._categories: SortedSet = SortedSet()
         self.bound_inf = 0.0
         self.bound_sup = 0.0
 
+        self.best_window_size = np.inf
+        # Default best window size. Re-measure it with self.measure_best_window_size
+
     @classmethod
     def from_csv(cls,
                  path: Union[str, Path],
@@ -194,6 +195,15 @@ class Continuum:
             continuum.add_annotation(uri, annot)
         return continuum
 
+    def copy_flush(self) -> 'Continuum':
+        """
+        Returns a copy of the continuum without any annotators/annotations, but with every other information
+        """
+        continuum = Continuum(self.uri)
+        continuum.bound_inf, continuum.bound_sup = self.bound_inf, self.bound_sup
+        continuum.best_window_size = self.best_window_size
+        return continuum
+
     def copy(self) -> 'Continuum':
         """
         Makes a copy of the current continuum.
@@ -205,6 +215,7 @@ class Continuum:
         continuum = Continuum(self.uri)
         continuum._annotations = deepcopy(self._annotations)
         continuum.bound_inf, continuum.bound_sup = self.bound_inf, self.bound_sup
+        continuum.best_window_size = self.best_window_size
         return continuum
 
     def __bool__(self):
@@ -228,14 +239,13 @@ class Continuum:
     @property
     def categories(self) -> SortedSet:
         """Returns the (alphabetically) sorted set of all the continuum's annotations's categories."""
-        return SortedSet(unit.annotation for _, unit in self
-                         if unit.annotation is not None)
+        return self._categories
 
     @property
     def category_weights(self) -> SortedDict:
-        """from .notebook import show_continuum
-        Returns a dictionnary where the keys are the categories in the continuum, and a key's value
-        is the proportion of occurence of the category in the continuum.
+        """
+        Returns a dictionary where the keys are the categories in the continuum, and a key's value
+        is the proportion of occurrence of the category in the continuum.
         """
         weights = SortedDict()
         nb_units = 0
@@ -251,7 +261,8 @@ class Continuum:
 
     @property
     def bounds(self) -> Tuple[float, float]:
-        """Bounds of the continuum. Initated as (0, 0), they grow as annotations are added."""
+        """Bounds of the continuum. Initially defined as (0, 0),
+        they grow as annotations are added."""
         return self.bound_inf, self.bound_sup
 
     @property
@@ -305,7 +316,7 @@ class Continuum:
 
         if annotator not in self._annotations:
             self._annotations[annotator] = SortedSet()
-
+        self._categories.add(annotation)
         self._annotations[annotator].add(Unit(segment, annotation))
         self.bound_inf = min(self.bound_inf, segment.start)
         self.bound_sup = max(self.bound_sup, segment.end)
@@ -486,6 +497,7 @@ class Continuum:
     def iter_annotator(self, annotator: Annotator) -> Generator[Unit, None, None]:
         """
         Iterates over the annotations of the given annotator.
+
         Raises
         ------
         KeyError
@@ -524,6 +536,121 @@ class Continuum:
         """
         return iter(self._annotations[annotator])
 
+    def get_first_window(self, dissimilarity: AbstractDissimilarity, w: int = 1) -> Tuple['Continuum', float]:
+        """
+        Returns a tuple (continuum, x_limit), where :
+            - Before x_limit, there are the (w * nb_annotators) leftmost annotations
+              of the continuum.
+            - After x_limit, there are (approximately) all the annotations from the continuum
+              that have a dissimilarity lower than (delta_empty * nb_annotators) with the annotations
+              before x_limit.
+        """
+        # Everything is converted to zippable lists. This is necessary for this method to have
+        # a simple and known complexity for choosing the most advantageous window size.
+        annotators = list(self.annotators)
+        annotations = list(self._annotations.values())
+        sizes = list(map(len, annotations))
+        indexes = [0] * len(annotators)  # Indexes for "advancing" homogeneously in the continuum
+
+        smallest_unit = Unit(Segment(-np.inf, -np.inf), None)
+
+        window = Continuum()
+        for annotator in annotators:
+            window.add_annotator(annotator)
+        taken_units = 0
+        rightmost_unit = smallest_unit
+        to_take = min(float(np.sum(sizes)), w * self.num_annotators)
+        while taken_units < to_take:  # At least (nb_annotators * w) units
+
+            x_limit = np.inf
+            for units, index, size in zip(annotations, indexes, sizes):  # Taking the rightmost unit not already taken
+                if index >= size:  # All annotations have been consumed
+                    continue
+                unit = units[index]
+                x_limit = min(x_limit, unit.segment.end)
+
+            for i, (annotator, units, index, size) in enumerate(zip(annotators, annotations, indexes, sizes)):
+                if index >= size:  # All annotations have been consumed
+                    continue
+                unit = units[index]  # Adding the units before x_limit. This will take between 1 and nb_annotator units.
+                if unit.segment.end <= x_limit:
+                    window.add(annotator, unit.segment, unit.annotation)
+                    rightmost_unit = max(unit, rightmost_unit)  # Rightmost taken unit is kept
+                    taken_units += 1                            # for selection of additionnal units.
+                    indexes[i] += 1
+
+        x_limit = window.bound_sup
+
+        # Now we add the additionnal annotations, "reachable" from those already selected.
+        for annotator, units, index, size in zip(annotators, annotations, indexes, sizes):
+            while index < size:
+                unit = units[index]
+                if dissimilarity.d(rightmost_unit, unit) > dissimilarity.delta_empty * self.num_annotators:
+                    break
+                window.add(annotator, unit.segment, unit.annotation)
+                index += 1
+        return window, x_limit
+
+    def get_fast_alignment(self, dissimilarity: AbstractDissimilarity, window_size: int) -> 'Alignment':
+        """Returns an 'approximation' of the best alignment (Very likely to be the actual best alignment for
+         continua with limited overlapping)"""
+        from .alignment import Alignment
+        copy = self.copy()
+        unitary_alignments = []
+        disorders = []
+
+        while copy:
+            window, x_limit = copy.get_first_window(dissimilarity, window_size)
+            # Window contains each annotator's first annotations
+            # We retain only the leftmost unitary alignment in the best alignment of the window,
+            # as it is the most likely to be in the global best alignment
+            best_alignment = window.get_best_alignment(dissimilarity)
+            for chosen in best_alignment.take_until_limit(x_limit):
+                unitary_alignments.append(chosen)
+                disorders.append(chosen.disorder)
+                for annotator, unit in chosen.n_tuple:
+                    if unit is not None:
+                        copy.remove(annotator, unit)  # Now we remove the units from the chosen alignment.
+        return Alignment(unitary_alignments,
+                         self,
+                         check_validity=False,  # Validity has been thoroughly tested
+                         disorder=np.sum(disorders) / self.avg_num_annotations_per_annotator)
+
+    def measure_best_window_size(self, dissimilarity: AbstractDissimilarity):
+        """
+        Sets the best window size for computing the fast-gamma of this continuum, by using the
+        sampling the computing complexity function.
+        """
+        smallest_window, _ = self.get_first_window(dissimilarity, 1)
+        smallest_window.get_best_alignment(dissimilarity)
+
+        s = smallest_window.max_num_annotations_per_annotator
+        n = int(self.avg_num_annotations_per_annotator)
+        p = int(self.num_annotators)
+
+        window_sizes = np.arange(1, max(2, self.max_num_annotations_per_annotator))
+        numba_factor = 1/20
+
+        def f(w):
+            return (n * p +  # Copying the continuum
+                    + ((n - w) * p / 2 + 2 * p + (w + s * p) * p  # getting first window
+                        + (n - w) * p / 2 + numba_factor * (w + s) ** p  # getting best alignment
+                        + (w + s) * p * np.log2((w + s) * p) + w * p  # getting the w leftmost alignments & adding them
+                       ) * (n / w))
+
+        # adding the log factorial corresponds to the emptying-the-continuum-unit-by-unit time.
+        logfactorials = np.log2(window_sizes)
+        for i in range(1, len(logfactorials)):
+            logfactorials[i] += logfactorials[i - 1]
+
+        times = f(window_sizes) + p * logfactorials
+
+        min_index = np.argmin(times)
+        if times[min_index] < n * p + numba_factor * n**p:  # Check is fast-gamma is advantageous compared to gamma
+            self.best_window_size = window_sizes[min_index]
+        else:
+            logging.warning("Fast-gamma disadvantageous, using normal gamma.")
+
     def get_best_alignment(self, dissimilarity: AbstractDissimilarity) -> 'Alignment':
         """
         Returns the best alignment of the continuum for the given dissimilarity. This alignment comes
@@ -540,25 +667,12 @@ class Continuum:
         assert len(self.annotators) >= 2 and self, "Disorder cannot be computed with less than two annotators, or " \
                                                    "without annotations."
 
-        disorder_args = dissimilarity.build_args(self)
-
         nb_unit_per_annot = []
         for annotator, arr in self._annotations.items():
             # assert len(arr) > 0, f"Disorder cannot be computed because annotator {annotator} has no annotations."
             nb_unit_per_annot.append(len(arr) + 1)
 
-        all_disorders = []
-        all_valid_tuples = []
-        for tuples_batch in chunked_cartesian_product(nb_unit_per_annot, CHUNK_SIZE):
-            batch_disorders = dissimilarity(tuples_batch, *disorder_args)
-            # Property section 5.1.1 to reduce initial complexity
-            valid_disorders_ids, = np.where(batch_disorders <= self.num_annotators * dissimilarity.delta_empty)
-
-            all_disorders.append(batch_disorders[valid_disorders_ids])
-            all_valid_tuples.append(tuples_batch[valid_disorders_ids])
-
-        disorders = np.concatenate(all_disorders)
-        possible_unitary_alignments = np.concatenate(all_valid_tuples)
+        disorders, possible_unitary_alignments = dissimilarity.valid_alignments(self)
 
         # Definition of the integer linear program
         n = len(disorders)
@@ -619,7 +733,8 @@ class Continuum:
                       n_samples: int = 30,
                       precision_level: Optional[Union[float, PrecisionLevel]] = None,
                       ground_truth_annotators: Optional[SortedSet] = None,
-                      sampler: 'AbstractContinuumSampler' = None) -> 'GammaResults':
+                      sampler: 'AbstractContinuumSampler' = None,
+                      fast: bool = False) -> 'GammaResults':
         """
 
         Parameters
@@ -641,65 +756,73 @@ class Continuum:
         sampler: AbstractContinuumSampler
             Sampler object, which implements a sampling strategy for creating random continuua used
             to calculate the expected disorder. If not set, defaults to the Statistical continuum sampler
+        fast:
+            Sets the algorithm to the much faster fast-gamma. It's supposed to be less precise than the "canonical"
+            algorithm from Mathet 2015, but usually isn't.
+            Performance gains and precision are explained in the Performance section of the documentation.
         """
         from .dissimilarity import CombinedCategoricalDissimilarity
         if dissimilarity is None:
-            dissimilarity = CombinedCategoricalDissimilarity(self.categories)
+            dissimilarity = CombinedCategoricalDissimilarity()
 
         if sampler is None:
             from .sampler import StatisticalContinuumSampler
             sampler = StatisticalContinuumSampler()
         sampler.init_sampling(self, ground_truth_annotators)
 
-        # Multiprocessed computation of sample disorder
-        p = Pool()
-        # computation of best alignment in advance
-        best_alignment_task = p.apply_async(_compute_best_alignment_job,
-                                            (dissimilarity, self,))
-        result_pool = [
-            # Step one : computing the disorders of a batch of random samples from the continuum (done in parallel)
-            p.apply_async(_compute_best_alignment_job,
-                          (dissimilarity, sampler.sample_from_continuum,))
-            for _ in range(n_samples)
-        ]
-        chance_best_alignments: List[Alignment] = []
-        chance_disorders: List[float] = []
+        if fast:
+            job = _compute_fast_alignment_job
+            self.measure_best_window_size(dissimilarity)
+        else:
+            job = _compute_best_alignment_job
 
-        best_alignment = best_alignment_task.get()
-        logging.info("Best alignment obtained...")
+        # Multithreaded computation of sample disorder
+        with ThreadPoolExecutor(max_workers=os.cpu_count()) as p:
+            # computation of best alignment in advance
+            best_alignment_task = p.submit(job,
+                                           *(dissimilarity, self))
+            best_alignment = best_alignment_task.result()
+            logging.info("Best alignment obtained...")
 
-        logging.info(f"Starting computation for a batch of {n_samples} random samples...")
-        for i, result in enumerate(result_pool):
-            chance_best_alignments.append(result.get())
-            logging.info(f"finished computation of random sample dissimilarity {i + 1}/{n_samples}")
-            chance_disorders.append(chance_best_alignments[-1].disorder)
-        logging.info("done.")
+            result_pool = [
+                # Step one : computing the disorders of a batch of random samples from the continuum (done in parallel)
+                p.submit(job,
+                         *(dissimilarity, sampler.sample_from_continuum))
+                for _ in range(n_samples)
+            ]
+            chance_best_alignments: List[Alignment] = []
+            chance_disorders: List[float] = []
 
-        if precision_level is not None:
-            if isinstance(precision_level, str):
-                precision_level = PRECISION_LEVEL[precision_level]
-            assert 0 < precision_level < 1.0
-            # If the variation of the disorders of the samples si too high, others are generated.
-            # taken from subsection 5.3 of the original paper
-            # confidence at 95%, i.e., 1.96
-            variation_coeff = np.std(chance_disorders) / np.mean(chance_disorders)
-            confidence = 1.96
-            required_samples = np.ceil((variation_coeff * confidence / precision_level) ** 2).astype(np.int32)
-            if required_samples > n_samples:
-                logging.info(f"Computing second batch of {required_samples - n_samples} "
-                             f"because variation was too high.")
-                result_pool = [
-                    p.apply_async(_compute_best_alignment_job,
-                                  (dissimilarity, sampler.sample_from_continuum,))
-                    for _ in range(required_samples - n_samples)
-                ]
-                for i, result in enumerate(result_pool):
-                    chance_best_alignments.append(result.get())
-                    logging.info(f"finished computation of additionnal random sample dissimilarity "
-                                 f"{i + 1}/{required_samples - n_samples}")
-                logging.info("done.")
+            logging.info(f"Starting computation for a batch of {n_samples} random samples...")
+            for i, result in enumerate(result_pool):
+                chance_best_alignments.append(result.result())
+                logging.info(f"finished computation of random sample dissimilarity {i + 1}/{n_samples}")
+                chance_disorders.append(chance_best_alignments[-1].disorder)
+            logging.info("done.")
 
-        p.close()
+            if precision_level is not None:
+                if isinstance(precision_level, str):
+                    precision_level = PRECISION_LEVEL[precision_level]
+                assert 0 < precision_level < 1.0
+                # If the variation of the disorders of the samples si too high, others are generated.
+                # taken from subsection 5.3 of the original paper
+                # confidence at 95%, i.e., 1.96
+                variation_coeff = np.std(chance_disorders) / np.mean(chance_disorders)
+                confidence = 1.96
+                required_samples = np.ceil((variation_coeff * confidence / precision_level) ** 2).astype(np.int32)
+                if required_samples > n_samples:
+                    logging.info(f"Computing second batch of {required_samples - n_samples} "
+                                 f"because variation was too high.")
+                    result_pool = [
+                        p.submit(job,
+                                 *(dissimilarity, sampler.sample_from_continuum))
+                        for _ in range(required_samples - n_samples)
+                    ]
+                    for i, result in enumerate(result_pool):
+                        chance_best_alignments.append(result.result())
+                        logging.info(f"finished computation of additionnal random sample dissimilarity "
+                                     f"{i + 1}/{required_samples - n_samples}")
+                    logging.info("done.")
 
         return GammaResults(
             best_alignment=best_alignment,
@@ -756,41 +879,12 @@ class GammaResults:
         observed disagreement."""
         return self.best_alignment.disorder
 
-    @property
-    def observed_cat_disorder(self) -> float:
-        """Observed disorder for gamma-cat (disorder of the best alignment)"""
-        return self.best_alignment.gamma_k_disorder(self.dissimilarity, None)
-
-    def observed_k_disorder(self, category: str) -> float:
-        """Observed disorder for gamma-k of the given category (disorder of best alignment)"""
-        return self.best_alignment.gamma_k_disorder(self.dissimilarity, category)
-
     @property
     def expected_disorder(self) -> float:
         """Returns the expected disagreement for computed random samples, i.e.,
         the mean of the sampled continuua's disorders"""
         return float(np.mean([align.disorder for align in self.chance_alignments]))
 
-    @property
-    def expected_cat_disorder(self) -> float:
-        """
-        Returns the expected disagreement (as defined for gamma-cat)
-        using the same random samples' best alignments
-        as for gamma (the mean of the sampled continuua's gamma-cat disorders)
-        """
-        return float(np.mean(list(filter((lambda x: x is not np.NaN),
-                                         (align.gamma_k_disorder(self.dissimilarity, None)
-                                          for align in self.chance_alignments)))))
-
-    def expected_k_disorder(self, category: str) -> float:
-        """
-        Returns the expected disagreement (as defined for gamma-k)
-        using the same random samples' best alignments
-        as for gamma (the mean of the sampled continuua's gamma-k disorders)
-        """
-        return float(np.mean(list(filter((lambda x: x is not np.NaN),
-                                         (align.gamma_k_disorder(self.dissimilarity, category)
-                                          for align in self.chance_alignments)))))
 
     @property
     def approx_gamma_range(self):
@@ -815,17 +909,40 @@ class GammaResults:
     @property
     def gamma_cat(self) -> float:
         """Returns the gamma-cat value"""
-        observed_cat_disorder = self.observed_cat_disorder
-        if observed_cat_disorder == 0:
-            return 1
-        return 1 - observed_cat_disorder / self.expected_cat_disorder
+        with ThreadPoolExecutor(max_workers=os.cpu_count()) as p:
+
+            observed_disorder_job = p.submit(_compute_gamma_k_job,
+                                             *(self.dissimilarity, self.best_alignment, None))
+
+            chance_disorders_jobs = [
+                p.submit(_compute_gamma_k_job,
+                         *(self.dissimilarity, alignment, None))
+                for alignment in self.chance_alignments
+            ]
+            observed_disorder = observed_disorder_job.result()
+            if observed_disorder == 0:
+                return 1
+            expected_disorder = float(np.mean(np.array([job_res.result() for job_res in chance_disorders_jobs])))
+
+        return 1 - observed_disorder / expected_disorder
 
     def gamma_k(self, category: str) -> float:
         """Returns the gamma-k value for the given category"""
-        observed_k_disorder = self.observed_k_disorder(category)
-        if observed_k_disorder == 0:
-            return 1
-        return 1 - observed_k_disorder / self.expected_k_disorder(category)
+        with ThreadPoolExecutor(max_workers=os.cpu_count()) as p:
+            observed_disorder_job = p.submit(_compute_gamma_k_job,
+                                             *(self.dissimilarity, self.best_alignment, category))
+
+            chance_disorders_jobs = [
+                p.submit(_compute_gamma_k_job,
+                         *(self.dissimilarity, alignment, category))
+                for alignment in self.chance_alignments
+            ]
+            observed_disorder = observed_disorder_job.result()
+            if observed_disorder == 0:
+                return 1
+            expected_disorder = float(np.mean(np.array([job_res.result() for job_res in chance_disorders_jobs])))
+
+        return 1 - observed_disorder / expected_disorder
 
 
 def _compute_best_alignment_job(dissimilarity: AbstractDissimilarity,
@@ -835,3 +952,19 @@ def _compute_best_alignment_job(dissimilarity: AbstractDissimilarity,
     using the given dissimilarity.
     """
     return continuum.get_best_alignment(dissimilarity)
+
+
+def _compute_fast_alignment_job(dissimilarity: AbstractDissimilarity,
+                                continuum: Continuum):
+    """
+    Function used to launch a multiprocessed job for calculating an approximation of
+    the best aligment of a continuum, using the given dissimilarity.
+    """
+    if continuum.best_window_size == np.inf:  # window size is set to infinity when normal gamma is better.
+        return continuum.get_best_alignment(dissimilarity)
+    return continuum.get_fast_alignment(dissimilarity, continuum.best_window_size)
+
+def _compute_gamma_k_job(dissimilarity: AbstractDissimilarity,
+                         alignment: 'Alignment',
+                         category: Optional[str]):
+    return alignment.gamma_k_disorder(dissimilarity, category)
\ No newline at end of file
